@@ -7,10 +7,16 @@ import types
 import itertools
 import logging
 import ast
+import hashlib
+import codecs
+try:
+    import cPickle as pickle
+except ImportError:
+    import pickle
 
 from ply import lex, yacc
 
-from . import PLYPLUS_DIR, grammar_parser
+from . import __version__, PLYPLUS_DIR, grammar_parser
 from .utils import StringTypes, StringType
 
 from .strees import STree, SVisitor, STransformer, is_stree, SVisitor_Recurse, Str
@@ -68,6 +74,10 @@ from .strees import STree, SVisitor, STransformer, is_stree, SVisitor_Recurse, S
 #DONE: Better error handling (choose between prints and raising exception, setting threshold, etc.)
 #
 
+grammar_logger = logging.getLogger('Grammar')
+grammar_logger.setLevel(logging.ERROR)
+
+
 _TOKEN_NAMES = {
     ':' : 'COLON',
     ',' : 'COMMA',
@@ -154,8 +164,7 @@ class CollectTokenDefs_Visitor(SVisitor):
     def tokendef(self, tree):
         self.tokendefs[ tree.tail[0] ] = tree
 
-    def fragmentdef(self, tree):
-        self.tokendefs[ tree.tail[0] ] = tree
+    fragmentdef = tokendef
 
 def _unescape_token_def(token_def):
     assert token_def[0] == "'" == token_def[-1]
@@ -467,7 +476,7 @@ class TokValue(Str):
 class LexerWrapper(object):
     def __init__(self, lexer, newline_tokens_names, newline_char='\n', ignore_token_names=()):
         self.lexer = lexer
-        self.newline_tokens_names = set(newline_tokens_names)
+        self.newline_tokens_names = frozenset(newline_tokens_names)
         self.ignore_token_names = ignore_token_names
         self.newline_char = newline_char
 
@@ -518,9 +527,9 @@ class LexerWrapper(object):
 
     def _handle_newlines(self, t):
         newlines = t.value.count(self.newline_char)
-        self.lineno += newlines
 
         if newlines:
+            self.lineno += newlines
             self._lexer_pos_of_start_column = t.lexpos + t.value.rindex(self.newline_char)
 
 
@@ -546,11 +555,27 @@ class Grammar(object):
 
         assert isinstance(grammar, StringTypes)
 
+        cache_grammar = options.pop('cache_grammar', False)
+        if cache_grammar:
+            plyplus_cache_filename = PLYPLUS_DIR + '/%s-%s-%s.plyplus' % (tab_filename, hashlib.sha256(grammar).hexdigest(), __version__)
+            if os.path.exists(plyplus_cache_filename):
+                with open(plyplus_cache_filename, 'rb') as f:
+                    self._grammar = pickle.load(f)
+            else:
+                self._grammar = self._create_grammar(grammar, source, tab_filename, options)
+
+                with open(plyplus_cache_filename, 'wb') as f:
+                    pickle.dump(self._grammar, f, pickle.HIGHEST_PROTOCOL)
+        else:
+            self._grammar = self._create_grammar(grammar, source, tab_filename, options)
+
+    @staticmethod
+    def _create_grammar(grammar, source, tab_filename, options):
         grammar_tree = grammar_parser.parse(grammar)
         if not grammar_tree:
             raise GrammarException("Parse Error: Could not create grammar")
 
-        self._grammar = _Grammar(grammar_tree, source, tab_filename, **options)
+        return _Grammar(grammar_tree, source, tab_filename, **options)
 
     def lex(self, text):
         return self._grammar.lex(text)
@@ -581,9 +606,6 @@ class _Grammar(object):
         self.lexer_postproc = None
         self._newline_value = '\n'
 
-        self.logger = logging.getLogger('Grammar')
-        self.logger.setLevel(logging.ERROR)
-
         # -- Build Grammar --
         self.subgrammars = {}
         ExtractSubgrammars_Visitor(source_name, tab_filename, self.options).visit(grammar_tree)
@@ -620,7 +642,7 @@ class _Grammar(object):
 
         # -- Build Parser --
         if not self.just_lex:
-            self.parser = yacc.yacc(module=self, debug=self.debug, tabmodule=tab_filename, errorlog=self.logger, outputdir=PLYPLUS_DIR)
+            self.parser = yacc.yacc(module=self, debug=self.debug, tabmodule=tab_filename, errorlog=grammar_logger, outputdir=PLYPLUS_DIR)
 
     def __repr__(self):
         return '<Grammar from %s, tab at %s>' % (self.source_name, self.tab_filename)
@@ -695,6 +717,7 @@ class _Grammar(object):
         # which is bad because our regexps are whitespace agnostic.
         # It also unescapes double backslashes, which messes up with the
         # regexp.
+
         token_value = token_value.replace('\\'*2, '\\'*4)
         # The equivalent whitespace escaping is:
         # token_value = token_value.replace(r'\n', r'\\n')
@@ -703,7 +726,7 @@ class _Grammar(object):
         # but for speed reasons, I ended-up with this ridiculus regexp:
         token_value = re.sub(r'(\\[nrf])', r'\\\1', token_value)
 
-        return token_value.decode('unicode-escape')
+        return codecs.getdecoder('unicode_escape')(token_value)[0]
 
     def _add_token_with_mods(self, name, defin):
         token_value, token_features = defin
@@ -775,7 +798,7 @@ class _Grammar(object):
 
         def p_rule(self, p):
             subtree = []
-            for child in p[1:]:
+            for child in p.__getslice__(1, None):
                 if isinstance(child, self.tree_class) and (
                            (                            child.head in self.rules_to_expand )
                         or (child.head == rule_name and child.head in self.rules_to_flatten)
@@ -791,7 +814,7 @@ class _Grammar(object):
 
             # Apply auto-filtering (remove 'punctuation' tokens)
             if self.auto_filter_tokens and len(subtree) != 1:
-                subtree = filter(is_stree, subtree)
+                subtree = list(filter(is_stree, subtree))
 
             if len(subtree) == 1 and (RuleMods.EXPAND in mods or RuleMods.EXPAND1 in mods):
                 # Self-expansion: only perform on EXPAND and EXPAND1 rules
