@@ -33,8 +33,6 @@ class WeakPickleMixin(object):
             if key in self.weak_attributes and val is not None:
                 setattr(self, key, ref(val))
 
-        self.__dict__['_cache'] = {}
-
 class Str(WeakPickleMixin, StringType): pass
 
 class STree(WeakPickleMixin, object):
@@ -83,6 +81,16 @@ class STree(WeakPickleMixin, object):
                 return
         raise ValueError("head not found: %s"%head)
 
+    def remove_kids_by_head(self, head):
+        removed = 0
+        for i, child in reversed(list(enumerate(self.tail))):
+            if is_stree(child) and child.head == head:
+                del self.tail[i]
+                removed += 1
+        if removed:
+            self.clear_cache()
+        return removed
+
     def remove_kid_by_id(self, child_id):
         for i, child in enumerate(self.tail):
             if id(child) == child_id:
@@ -91,6 +99,12 @@ class STree(WeakPickleMixin, object):
                 return
         raise ValueError("id not found: %s"%child_id)
 
+    def prune_by_head(self, head):
+        self.remove_kids_by_head(head)
+        for kid in self.tail:
+            if hasattr(kid, 'prune_by_head'):
+                kid.prune_by_head(head)
+
     def remove_from_parent(self):
         self.parent().remove_kid_by_id(id(self))
         self.parent = None
@@ -121,6 +135,12 @@ class STree(WeakPickleMixin, object):
         dict.pop('_cache', None)
         return dict
 
+    def __setstate__(self, data):
+        super(STree, self).__setstate__(data)
+
+        # Ensure we've got a clean cache
+        self.clear_cache()
+
     def __deepcopy__(self, memo):
         return type(self)(self.head, deepcopy(self.tail, memo))
 
@@ -155,6 +175,30 @@ class STree(WeakPickleMixin, object):
             self.parent = None
             self.index_in_parent = None
 
+    def calc_position(self):
+        for kid in self.tail:
+            if is_stree(kid):
+                kid.calc_position()
+            else:
+                try:
+                    kid.min_line = kid.max_line = kid.line
+                    kid.min_col = kid.max_col = kid.column
+                except AttributeError:
+                    kid.min_line = kid.min_col = kid.max_line = kid.max_col = None
+
+        try:
+            first = next(x for x in self.tail if x.min_line is not None)
+            last = next(x for x in reversed(self.tail) if x.max_line is not None)
+        except StopIteration:
+            self.min_line = self.min_col = self.max_line = self.max_col = None
+            return
+
+        self.min_line = first.min_line
+        self.min_col = first.min_col
+        self.max_line = last.max_line
+        self.max_col = last.max_col
+
+
     def calc_depth(self, depth=0):
         self.depth = depth
         for kid in self.tail:
